@@ -1,5 +1,5 @@
 ; libctl: flexible Guile-based control files for scientific software 
-; Copyright (C) 1998, 1999, 2000, 2001, 2002, Steven G. Johnson
+; Copyright (C) 1998, 1999, 2000, 2001, 2002, 2006, Steven G. Johnson
 ;
 ; This library is free software; you can redistribute it and/or
 ; modify it under the terms of the GNU Lesser General Public
@@ -32,7 +32,9 @@
 (define (object-type-names object) (car object))
 (define (object-property-values object) (cdr object))
 (define (object-member? type-name object)
-  (and (pair? object) (member type-name (object-type-names object))))
+  (and (pair? object) 
+       (list? (object-type-names object))
+       (member type-name (object-type-names object))))
 (define (object-property-value object property-name)
   (assoc-ref (object-property-values object) property-name))
 (define (object-memberp type-name)
@@ -85,6 +87,10 @@
   (and (pair? type-name) (eq? (car type-name) 'list)))
 (define (list-el-type-name type-name) (cdr type-name))
 
+(define exported-type-list '())
+(define (export-type type-name)
+  (set! exported-type-list (cons type-name exported-type-list)))
+
 (define (make-type-descriptor kind name name-str predicate)
   (list kind name name-str predicate))
 (define type-descriptor-kind first)
@@ -121,6 +127,12 @@
    ((list-type-name? type-name) (make-list-type-descriptor type-name))
    (else (error "unknown type" type-name))))
 
+(define (primitive-type? type-name)
+  (or (eq? type-name 'number)
+      (eq? type-name 'integer)
+      (eq? type-name 'function)
+      (eq? type-name 'SCM)))
+
 (define (type-string type-name)
   (let ((desc (get-type-descriptor type-name)))
     (cond
@@ -168,6 +180,10 @@
 (define (class-type-name class) (first class))
 (define (class-parent class) (second class))
 (define (class-properties class) (third class))
+(define (class-properties-all class)
+  (append (class-properties class)
+	  (let ((parent (class-parent class)))
+	    (if parent (class-properties parent) '()))))
 (define (class-member? type-name class)
   (if (list? class)
       (or (eq? type-name (class-type-name class))
