@@ -1,5 +1,5 @@
 /* libctl: flexible Guile-based control files for scientific software 
- * Copyright (C) 1998, 1999, 2000, 2001, 2002, Steven G. Johnson
+ * Copyright (C) 1998, 1999, 2000, 2001, 2002, 2006, Steven G. Johnson
  *
  * This file may be used without restriction.  It is in the public
  * domain, and is NOT restricted by the terms of any GNU license.
@@ -32,7 +32,14 @@
 #include <string.h>
 #include <guile/gh.h>
 
-#include <ctl-io.h>
+/* for basename and dirname functions */
+#include <libgen.h>
+
+#include "ctl-io.h"
+
+#ifdef CXX_CTL_IO
+using namespace ctlio;
+#endif
 
 /* define a global "verbose" variable set by the --verbose command-line opt. */
 int verbose = 0;
@@ -102,6 +109,37 @@ int handle_args(int argc, char *argv[],
 
 /**************************************************************************/
 
+static int exists(const char *fname)
+{
+     FILE *f = fopen(fname, "r");
+     if (f) {
+	  fclose(f);
+	  return 1;
+     }
+     return 0;
+}
+
+static char *make_name(const char *for_dir, const char *for_base)
+{
+     char *dir0, *dir, *base0, *base, *name = 0;
+     dir0 = (char *) malloc(sizeof(char) * (strlen(for_dir) + 1));
+     base0 = (char *) malloc(sizeof(char) * (strlen(for_base) + 1));
+     strcpy(dir0, for_dir); dir = dirname(dir0);
+     if (strlen(dir)) {
+	  strcpy(base0, for_base); base = basename(base0);
+	  name = (char *) malloc(sizeof(char) * (strlen(dir) + 1 +
+						 strlen(base) + 1));
+	  strcpy(name, dir);
+	  strcat(name, "/");
+	  strcat(name, base);
+	  free(base0);
+     }
+     free(dir0);
+     return name;
+}
+
+/**************************************************************************/
+
 #ifdef HAVE_CTL_HOOKS
 static int ctl_stop_hook_called = 0;
 
@@ -134,7 +172,16 @@ void main_entry(int argc, char *argv[])
   gh_new_procedure ("write-output-vars", write_output_vars, 0, 0, 0);
 
   /* Export the subplex minimization routine: */
-  gh_new_procedure ("subplex", subplex_scm, 7, 0, 0);
+  gh_new_procedure ("subplex", (SCM (*)(void)) subplex_scm, 7, 0, 0);
+
+  /* Export the adaptive integration routines: */
+  gh_new_procedure ("adaptive-integration", 
+		    (SCM (*)(void)) adaptive_integration_scm, 6, 0, 0);
+
+#ifdef CTL_HAS_COMPLEX_INTEGRATION
+  gh_new_procedure ("cadaptive-integration", 
+		    (SCM (*)(void)) cadaptive_integration_scm, 6, 0, 0);
+#endif
 
 #ifdef HAVE_CTL_EXPORT_HOOK
   ctl_export_hook();
@@ -152,14 +199,30 @@ void main_entry(int argc, char *argv[])
 
   i = handle_args(argc, argv, &spec_file_loaded, &continue_run);
 
+  {
+       char definestr[] = "(define verbose? false)";
+       strcpy(definestr, "(define verbose? ");
+       strcat(definestr, verbose ? "true)" : "false)");
+       gh_eval_str(definestr);
+  }
+
   if (!continue_run)
        goto done;
 
   /* load the specification file if it was given at compile time,
      and if it wasn't specified on the command-line: */
 #ifdef SPEC_SCM
-  if (!spec_file_loaded)
-       ctl_include(SPEC_SCM);
+  if (!spec_file_loaded) {
+       /* try first to load it in the program directory if it
+	  was specified explicitly (e.g. "./foo"), for cases
+	  where we are running a program that has not been installed */
+       char *spec_name = make_name(argv[0], SPEC_SCM);
+       if (spec_name && exists(spec_name))
+	    ctl_include(spec_name);
+       else
+	    ctl_include(SPEC_SCM);
+       free(spec_name);
+  }
 #endif
 
   /* define any variables and load any scheme files specified on the
