@@ -1,5 +1,5 @@
 ; libctl: flexible Guile-based control files for scientific software 
-; Copyright (C) 1998, 1999, 2000, 2001, 2002, Steven G. Johnson
+; Copyright (C) 1998, 1999, 2000, 2001, 2002, 2006, Steven G. Johnson
 ;
 ; This library is free software; you can redistribute it and/or
 ; modify it under the terms of the GNU Lesser General Public
@@ -33,6 +33,11 @@
 
 ; ***************************************************************************
 
+(define cxx false) ; set to true for C++ output (c.f. gen-ctl-io --cxx)
+(define namespace "ctlio")
+(define (ns0) (ns namespace))
+(define (ns namespace) (if cxx (string-append namespace "::") ""))
+
 (define (c-identifier s)
   (list->string (map (lambda (c) 
 		       (if (or (eq? c #\-) (eq? c #\space))
@@ -58,10 +63,44 @@
 	(set! declared-type-names (cons (c-type-string type-name)
 					declared-type-names)))))
 
-(define (c-var-decl var-name var-type-name)
-  (print (c-type-string var-type-name) " " 
+(define (only-list-types type-names)
+  (list-transform-positive type-names list-type-name?))
+
+(define (c-var-decl' var-name var-type-name ns)
+  (print (c-type-string var-type-name) " " ns
 		(symbol->c-identifier var-name) ";\n"))
 
+(define (c-var-decl var-name var-type-name)
+  (c-var-decl' var-name var-type-name ""))
+
+; ***************************************************************************
+; use new/delete for C++, malloc/free for C
+
+(define (free . vars)
+  (let ((var (apply string-append vars)))
+    (if cxx 
+	(string-append "delete[] (" var ")")
+	(string-append "free(" var ")"))))
+      
+(define (free1 . vars)
+  (let ((var (apply string-append vars)))
+    (if cxx 
+	(string-append "delete (" var ")")
+	(string-append "free(" var ")"))))
+
+(define (malloc tname . nums)
+  (let ((num (apply string-append nums)))
+    (if cxx
+	(string-append "(new " tname "[" num "])")
+	(string-append 
+	 "((" tname " *) malloc(sizeof(" tname ") * (" num ")))"))))
+
+(define (malloc1 . tnames)
+  (let ((tname (apply string-append tnames)))
+    (if cxx
+	(string-append "(new " tname ")")
+	(string-append "((" tname " *) malloc(sizeof(" tname ")))"))))
+
 ; ***************************************************************************
 
 (define (find-direct-subclasses class)
@@ -71,9 +110,17 @@
 (define (class-identifier class)
   (symbol->c-identifier (class-type-name class)))
 
-(define (class-enum-name class)
+(define (class-enum-name0 class)
   (string-upcase (class-identifier class)))
 
+(define (class-enum-name class)
+  (string-append (ns (class-identifier (class-parent class)))
+		 (class-enum-name0 class)))
+
+(define (class-self-enum-name class)
+  (string-append (ns (class-identifier class))
+		 (class-enum-name0 class) "_SELF"))
+
 (define (c-class-decl class)
   (for-each (compose declare-type-name property-type-name)
 	    (class-properties class))
@@ -86,8 +133,8 @@
   (let ((subclasses (find-direct-subclasses class)))
     (if (not (null? subclasses))
 	(begin
-	  (print "enum { " (class-enum-name class) "_SELF" )
-	  (for-each (lambda (sc) (print ", " (class-enum-name sc)))
+	  (print "enum { " (class-enum-name0 class) "_SELF")
+	  (for-each (lambda (sc) (print ", " (class-enum-name0 sc)))
 		    subclasses)
 	  (print " } which_subclass;\n")
 	  (print "union {\n")
@@ -97,7 +144,11 @@
 		       (class-identifier sc) "_data;\n"))
 		    subclasses)
 	  (print "} subclass;\n"))))
-  (print "} " (class-identifier class) ";\n\n"))
+  (print "} " (class-identifier class) ";\n")
+  (if (and (not (null? (find-direct-subclasses class)))
+	   (null? (class-properties class)))
+      (print "#define " (class-enum-name0 class) "_ABSTRACT 1\n"))
+  (print "\n"))
        
 (define (display-c-class-decls)
   (print "/******* Type declarations *******/\n\n")
@@ -106,7 +157,7 @@
 ; ***************************************************************************
 
 (define (declare-var var)
-  (c-var-decl (var-name var) (var-type-name var)))
+  (c-var-decl' (var-name var) (var-type-name var) (ns0)))
 (define (declare-extern-var var)
   (print "extern ")
   (c-var-decl (var-name var) (var-type-name var)))
@@ -118,17 +169,19 @@
 	(begin (print "/* " (var-name var) 
 			     " is both input and output */\n")))))
 
-(define (declare-var-types)
-  (let ((var-types (append
-		    (map var-type-name 
-			 (append (reverse input-var-list) 
-				 (reverse output-var-list)))
-		    (map external-function-return-type-name
-			 external-function-list)
-		    (fold-left append '()
-				(map external-function-arg-type-names
-				     external-function-list)))))
-    (for-each declare-type-name var-types)))
+(define (all-type-names)
+  (append
+   exported-type-list
+   (map var-type-name 
+	(append (reverse input-var-list) 
+		(reverse output-var-list)))
+   (map external-function-return-type-name
+	external-function-list)
+   (fold-left append '()
+	      (map external-function-arg-type-names
+		   external-function-list))))
+
+(define (declare-var-types) (for-each declare-type-name (all-type-names)))
 
 (define (declare-vars declarer)
   (print "/******* Input variables *******/\n")
@@ -188,11 +241,9 @@
     (print "int " index-name-str ";\n")
     (print c-var-name-str 
 		  ".num_items = list_length(" lo-name-str ");\n")
-    (print c-var-name-str ".items = (" 
-		  (c-type-string type-name)
-		  "*) malloc(sizeof("
-		  (c-type-string type-name) ") * "
-		  c-var-name-str ".num_items);\n")
+    (print c-var-name-str ".items = " 
+	   (malloc (c-type-string type-name) c-var-name-str ".num_items") 
+	   ";\n")
     (print "for (" index-name-str " = 0; " index-name-str " < "
 		  c-var-name-str ".num_items; " index-name-str "++) {\n")
     (input-value index-name-str 
@@ -206,14 +257,14 @@
   (string-append (symbol->c-identifier type-name)
 		 "_input"))
 
-(define (class-input-function-decl class)
-  (print "void " 
+(define (class-input-function-decl class ns)
+  (print "void " ns
 		(class-input-function-name (class-type-name class))
 		"(SCM so, "
 		(c-type-string (class-type-name class)) " *o)"))
 
 (define (class-input-function class)
-  (class-input-function-decl class)
+  (class-input-function-decl class (ns0))
   (print "\n{\n")
   (for-each
    (lambda (property)
@@ -229,10 +280,8 @@
        (print "if (object_is_member(\"" (class-type-name sc) 
 		     "\", so)) {\n")
        (print "o->which_subclass = " (class-enum-name sc) ";\n")
-       (print "o->subclass." (class-identifier sc)
-		     "_data = (" (class-identifier sc)
-		     "*) malloc(sizeof("
-		     (class-identifier sc) "));\n")
+       (print "o->subclass." (class-identifier sc) "_data = "
+	      (malloc1 (class-identifier sc)) ";\n")
        (print (class-input-function-name (class-type-name sc)) 
 		     "(so, o->subclass." (class-identifier sc) "_data);\n"
 		     "}\nelse "))
@@ -241,14 +290,14 @@
        (begin
 	 (print "\n")
 	 (print "o->which_subclass = " 
-		       (class-enum-name class) "_SELF;\n"))))
+		       (class-self-enum-name class) ";\n"))))
   (print "}\n\n"))
 
 (define (output-class-input-functions-header)
   (print "/******* class input function prototypes *******/\n\n")
   (for-each
    (lambda (class)
-     (print "extern ") (class-input-function-decl class)
+     (print "extern ") (class-input-function-decl class "")
      (print ";\n"))
    class-list)
   (print "\n"))
@@ -259,7 +308,7 @@
 
 (define (input-vars-function)
   (print "/******* read input variables *******/\n\n")
-  (print "SCM read_input_vars(void)\n")
+  (print "SCM " (ns0) "read_input_vars(void)\n")
   (print "{\n")
   (print "if (num_read_input_vars++) destroy_input_vars();\n")
   (for-each
@@ -275,6 +324,181 @@
 
 ; ***************************************************************************
 
+(define (copy-value c0-var-name-str c-var-name-str type-name)
+  (let ((desc (get-type-descriptor type-name)))
+    (cond
+     ((eq? (type-descriptor-kind desc) 'simple)
+      (print c-var-name-str " = " c0-var-name-str ";\n"))
+     ((eq? (type-descriptor-kind desc) 'object)
+      (print (class-copy-function-name type-name) "(&"
+	     c0-var-name-str
+	     ", &" c-var-name-str ");\n"))
+     ((eq? (type-descriptor-kind desc) 'uniform-list)
+      (copy-list c0-var-name-str c-var-name-str
+		 (list-el-type-name type-name))))))
+
+(define (copy-list c0-var-name-str c-var-name-str type-name)
+  (print "{\n")
+  (let ((index-name-str (string-append "i" list-temp-suffix))
+	(saved-list-temp-suffix list-temp-suffix))
+    (set! list-temp-suffix (string-append list-temp-suffix "_t"))
+    (print "int " index-name-str ";\n")
+    (print c-var-name-str ".num_items = "
+	   c0-var-name-str ".num_items;\n")
+    (print c-var-name-str ".items = " 
+	   (malloc (c-type-string type-name) c-var-name-str ".num_items")
+	   ";\n")
+    (print "for (" index-name-str " = 0; " index-name-str " < "
+		  c-var-name-str ".num_items; " index-name-str "++) {\n")
+    (copy-value (string-append c0-var-name-str ".items[" index-name-str "]")
+		(string-append c-var-name-str ".items[" index-name-str "]")
+		type-name)
+    (print "}\n")
+    (set! list-temp-suffix saved-list-temp-suffix))
+  (print "}\n"))
+
+(define (class-copy-function-name type-name)
+  (string-append (symbol->c-identifier type-name)
+		 "_copy"))
+
+(define (class-copy-function-decl class ns)
+  (print "void " ns
+		(class-copy-function-name (class-type-name class))
+		"(const " (c-type-string (class-type-name class)) " *o0,"
+		(c-type-string (class-type-name class)) " *o)"))
+
+(define (class-copy-function class)
+  (class-copy-function-decl class (ns0))
+  (print "\n{\n")
+  (for-each
+   (lambda (property)
+     (copy-value (string-append "o0->" (symbol->c-identifier 
+					(property-name property)))
+		 (string-append "o->" (symbol->c-identifier 
+				       (property-name property)))
+		 (property-type-name property)))
+   (class-properties class))
+  (let ((subclasses (find-direct-subclasses class)))
+    (for-each
+     (lambda (sc)
+       (print "if (o0->which_subclass == " (class-enum-name sc) ") {\n")
+       (print "o->which_subclass = " (class-enum-name sc) ";\n")
+       (print "o->subclass." (class-identifier sc) "_data = "
+	      (malloc1 (class-identifier sc)) ";\n")
+       (print (class-copy-function-name (class-type-name sc)) 
+	      "(o0->subclass." (class-identifier sc) 
+	      "_data, o->subclass." (class-identifier sc) "_data);\n"
+	      "}\nelse "))
+     subclasses)
+   (if (not (null? subclasses))
+       (begin
+	 (print "\n")
+	 (print "o->which_subclass = " 
+		       (class-self-enum-name class) ";\n"))))
+  (print "}\n\n"))
+
+(define (output-class-copy-functions-header)
+  (print "/******* class copy function prototypes *******/\n\n")
+  (for-each
+   (lambda (class)
+     (print "extern ") (class-copy-function-decl class "")
+     (print ";\n"))
+   class-list)
+  (print "\n"))
+
+(define (output-class-copy-functions-source)
+  (print "/******* class copy functions *******/\n\n")
+  (for-each class-copy-function class-list))
+
+; ***************************************************************************
+
+(define (equal-value c0-var-name-str c-var-name-str type-name)
+  (let ((desc (get-type-descriptor type-name)))
+    (cond
+     ((primitive-type? type-name)
+      (print "if (" c-var-name-str " != " c0-var-name-str ") return 0;\n"))
+     ((eq? type-name 'string)
+      (print "if (strcmp(" c-var-name-str ", " c0-var-name-str
+	     ")) return 0;\n"))
+     ((eq? (type-descriptor-kind desc) 'simple)
+      (print "if (!" type-name "_equal("
+	     c-var-name-str ", " c0-var-name-str ")) return 0;\n"))
+     ((eq? (type-descriptor-kind desc) 'object)
+      (print "if (!" (class-equal-function-name type-name) "(&"
+	     c0-var-name-str
+	     ", &" c-var-name-str ")) return 0;\n"))
+     ((eq? (type-descriptor-kind desc) 'uniform-list)
+      (equal-list c0-var-name-str c-var-name-str
+		 (list-el-type-name type-name))))))
+
+(define (equal-list c0-var-name-str c-var-name-str type-name)
+  (print "{\n")
+  (let ((index-name-str (string-append "i" list-temp-suffix))
+	(saved-list-temp-suffix list-temp-suffix))
+    (set! list-temp-suffix (string-append list-temp-suffix "_t"))
+    (print "int " index-name-str ";\n")
+    (print "if (" c-var-name-str ".num_items != "
+	   c0-var-name-str ".num_items) return 0;\n")
+    (print "for (" index-name-str " = 0; " index-name-str " < "
+	   c-var-name-str ".num_items; " index-name-str "++) {\n")
+    (equal-value (string-append c0-var-name-str ".items[" index-name-str "]")
+		 (string-append c-var-name-str ".items[" index-name-str "]")
+		 type-name)
+    (print "}\n")
+    (set! list-temp-suffix saved-list-temp-suffix))
+  (print "}\n"))
+
+(define (class-equal-function-name type-name)
+  (string-append (symbol->c-identifier type-name)
+		 "_equal"))
+
+(define (class-equal-function-decl class ns)
+  (print "boolean " ns
+		(class-equal-function-name (class-type-name class))
+		"(const " (c-type-string (class-type-name class)) " *o0, "
+		"const " (c-type-string (class-type-name class)) " *o)"))
+
+(define (class-equal-function class)
+  (class-equal-function-decl class (ns0))
+  (print "\n{\n")
+  (for-each
+   (lambda (property)
+     (equal-value (string-append "o0->" (symbol->c-identifier 
+					(property-name property)))
+		 (string-append "o->" (symbol->c-identifier 
+				       (property-name property)))
+		 (property-type-name property)))
+   (class-properties class))
+  (let ((subclasses (find-direct-subclasses class)))
+    (if (not (null? subclasses))
+	(print "if (o0->which_subclass != o->which_subclass) return 0;\n"))
+    (for-each
+     (lambda (sc)
+       (print "if (o0->which_subclass == " (class-enum-name sc) ") {\n")
+       (print "if (!" (class-equal-function-name (class-type-name sc)) 
+	      "(o0->subclass." (class-identifier sc) 
+	      "_data, o->subclass." (class-identifier sc) "_data)) return 0;\n"
+	      "}\nelse "))
+     subclasses)
+    (print ";\n"))
+  (print "return 1;\n")
+  (print "}\n\n"))
+
+(define (output-class-equal-functions-header)
+  (print "/******* class equal function prototypes *******/\n\n")
+  (for-each
+   (lambda (class)
+     (print "extern ") (class-equal-function-decl class "")
+     (print ";\n"))
+   class-list)
+  (print "\n"))
+
+(define (output-class-equal-functions-source)
+  (print "/******* class equal functions *******/\n\n")
+  (for-each class-equal-function class-list))
+
+; ***************************************************************************
+
 (define (export-object-value c-var-name-str type-name exporter)
   (error "object output variables are not yet supported. "
 	 type-name c-var-name-str))
@@ -317,7 +541,7 @@
 
 (define (output-vars-function)
   (print "/******* write output variables *******/\n\n")
-  (print "SCM write_output_vars(void)\n")
+  (print "SCM " (ns0) "write_output_vars(void)\n")
   (print "{\n")
   (print "num_write_output_vars++;\n")
   (for-each
@@ -337,7 +561,7 @@
   (let ((desc (get-type-descriptor type-name)))
     (cond 
      ((eq? type-name 'string)
-      (print "free(" var-str ");\n"))
+      (print (free var-str) ";\n"))
      ((eq? (type-descriptor-kind desc) 'uniform-list)
       (destroy-list var-str (list-el-type-name type-name)))
      ((eq? (type-descriptor-kind desc) 'object)
@@ -347,8 +571,8 @@
   (string-append (symbol->c-identifier type-name)
 		 "_destroy"))
 
-(define (class-destroy-function-decl class)
-  (print "void " 
+(define (class-destroy-function-decl class ns)
+  (print "void " ns
 		(class-destroy-function-name (class-type-name class))
 		"("
 		(c-type-string (class-type-name class)) " o)"))
@@ -365,7 +589,7 @@
 		    el-type-name)
        (print "}\n")
     (print "}\n")
-    (print "free(" var-str ".items);\n")
+    (print (free var-str ".items") ";\n")
     (set! list-temp-suffix saved-suffix)))
 
 (define (destroy-object var-str type-name)
@@ -377,7 +601,7 @@
 		 (property-type-name property)))
 
 (define (class-destroy-function class)
-  (class-destroy-function-decl class)
+  (class-destroy-function-decl class (ns0))
   (print "\n{\n")
   (for-each
    (lambda (property) (destroy-property "o." property))
@@ -389,7 +613,7 @@
        (destroy-object (string-append "*o.subclass." 
 				      (class-identifier sc) "_data")
 		       (class-type-name sc))
-       (print "free(o.subclass." (class-identifier sc) "_data);\n")
+       (print (free1 "o.subclass." (class-identifier sc) "_data") ";\n")
        (print "}\n")
        (print "else "))
      subclasses)
@@ -402,7 +626,7 @@
   (print "/******* class destruction function prototypes *******/\n\n")
   (for-each
    (lambda (class)
-     (print "extern ") (class-destroy-function-decl class)
+     (print "extern ") (class-destroy-function-decl class "")
      (print ";\n"))
    class-list)
   (print "\n"))
@@ -413,7 +637,7 @@
 
 (define (destroy-input-vars-function)
   (print "/******* destroy input variables *******/\n\n")
-  (print "SCM destroy_input_vars(void)\n")
+  (print "SCM " (ns0) "destroy_input_vars(void)\n")
   (print "{\n")
   (for-each
    (lambda (var)
@@ -426,7 +650,7 @@
 
 (define (destroy-output-vars-function)
   (print "/******* destroy output variables *******/\n\n")
-  (print "SCM destroy_output_vars(void)\n")
+  (print "SCM " (ns0) "destroy_output_vars(void)\n")
   (print "{\n")
   (for-each
    (lambda (var)
@@ -444,11 +668,11 @@
   (if (null? lst) '()
       (cons start-index (list->indices (cdr lst) (+ start-index 1)))))
 
-(define (declare-external-function external-function)
-  (print "SCM " (symbol->c-identifier
-		       (external-function-aux-name
-			(external-function-name external-function)))
-		"(")
+(define (declare-external-function external-function ns)
+  (print "SCM " ns (symbol->c-identifier
+		    (external-function-aux-name
+		     (external-function-name external-function)))
+	 "(")
   (for-each
    (lambda (argnum)
      (if (> argnum 0) (print ", "))
@@ -486,7 +710,7 @@
    (lambda (ef)
      (declare-external-c-function ef)
      (print "extern ")
-     (declare-external-function ef)
+     (declare-external-function ef "")
      (print ";\n\n"))
    external-function-list)
   (print "\nextern void export_external_functions(void);\n")
@@ -497,6 +721,7 @@
    "gh_new_procedure(\""
    (external-function-aux-name (external-function-name external-function))
    "\", "
+   "(SCM (*)(void)) "
    (symbol->c-identifier
     (external-function-aux-name (external-function-name external-function)))
    ", "
@@ -504,7 +729,7 @@
    ", 0, 0);\n"))
 
 (define (output-export-external-functions)
-  (print "void export_external_functions(void)\n")
+  (print "void " (ns0) "export_external_functions(void)\n")
   (print "{\n")
   (for-each output-external-function-export external-function-list)
   (print "}\n\n"))
@@ -519,7 +744,7 @@
 		 "_to_scm(" c-name-str ");"))
 
 (define (output-external-function external-function)
-  (declare-external-function external-function) (print "\n")
+  (declare-external-function external-function (ns0)) (print "\n")
   (print "{\n")
 
   (if (not (eq? (external-function-return-type-name external-function)
@@ -592,6 +817,39 @@
 
 ; ***************************************************************************
 
+(define (swig-type-header type-name)
+  (print "%typemap(guile,in) " 
+	 (if cxx (string-append namespace "::") "")
+	 (c-type-string type-name) " {\n")
+  (if cxx (print "using namespace " namespace ";\n"))
+  (input-value "$input" "$1" type-name get-c-local)
+  (print "}\n")
+  (if (and (not (eq? 'object (type-descriptor-kind 
+			      (get-type-descriptor type-name))))
+	   (or (not (list-type-name? type-name))
+	       (eq? 'simple (type-descriptor-kind 
+			     (get-type-descriptor 
+			      (list-el-type-name type-name))))))
+      (begin
+	(print "%typemap(guile,out) " 
+	       (if cxx (string-append namespace "::") "")
+	       (c-type-string type-name) " {\n")
+	(if cxx (print "using namespace " namespace ";\n"))
+	(output-value "$result" "$1" type-name set-c-local)
+	(destroy-c-var "$1" type-name)
+	(print "}\n")))
+  (print "\n")
+)
+
+(define (output-swig-header)
+  (print "%{\n#include \"ctl-io.h\"\n%}\n\n")
+  (print "/******* SWIG type-conversion mappings *******/\n\n")
+  (for-each swig-type-header
+	    (append (only-list-types (all-type-names))
+		    (map class-type-name class-list))))
+
+; ***************************************************************************
+
 (define (output-header)
   (display-c-class-decls)
   (declare-vars-header)
@@ -603,15 +861,19 @@
   (print "extern SCM destroy_output_vars(void);\n\n")
   (output-external-functions-header)
   (output-class-input-functions-header)
+  (output-class-copy-functions-header)
+  (output-class-equal-functions-header)
   (output-class-destruction-functions-header)
 )
 
 (define (output-source)
   (declare-vars-source)
   (print
-   "int num_read_input_vars = 0; /* # calls to read_input_vars */\n"
-   "int num_write_output_vars = 0; /* # calls to read_input_vars */\n\n")
+   "int " (ns0) "num_read_input_vars = 0; /* # calls to read_input_vars */\n"
+   "int " (ns0) "num_write_output_vars = 0; /* # calls to read_input_vars */\n\n")
   (output-class-input-functions-source)
+  (output-class-copy-functions-source)
+  (output-class-equal-functions-source)
   (output-class-destruction-functions-source)
   (input-vars-function)
   (output-vars-function)
